"""
contraints applied to callable objects
"""
from textx import get_children_of_type
from textx import get_location
from textx import textx_isinstance
from imports import check_function_import


class ArgumentMismatchError(TypeError):
    """raise this exception if arguments do not match the dummy variables"""
    def __init__(self, obj):
        msg = f'obj: {obj.name}, loc: {repr(get_location(obj))}'
        message = 'dummy variables do not match function arguments:\n' + msg
        super().__init__(message)


def check_function_definition(obj, metamodel):
    """check matching args and function_definition"""
    assert textx_isinstance(obj, metamodel['FunctionDefinition'])
    vrefs = get_children_of_type('VarReference', obj.expr)
    refs = [r.ref for r in vrefs if textx_isinstance(r.ref, metamodel['Dummy'])]
    refs_uniq_names = set(v.name for v in refs)
    args_names = [a.name for a in obj.args]
    args_uniq_names = set(args_names)
    assert len(args_names) == len(args_uniq_names)
    calls = get_children_of_type('FunctionCall', obj.expr)
    call_dummies = set(a.name for call in calls for a in call.func.args)
    assert not any(d in args_uniq_names for d in call_dummies)
    args_uniq_names.update(call_dummies)
    refs_uniq_names.update(call_dummies)
    try:
        assert len(refs_uniq_names) == len(args_uniq_names)
        assert list(sorted(refs_uniq_names)) == list(sorted(args_uniq_names))
    except AssertionError:
        print(len(refs_uniq_names), len(args_uniq_names))
        print(list(sorted(refs_uniq_names)), list(sorted(args_uniq_names)))
        raise ArgumentMismatchError(obj)


def check_function_call(obj, metamodel):
    """check a function call object"""
    if textx_isinstance(obj.func, metamodel['ObjectImport']):
        check_function_import(obj.func)  # has a valid imported function
    elif textx_isinstance(obj.func, metamodel['FunctionDefinition']):
        try:
            assert len(obj.params) == len(obj.func.args)
        except AssertionError:
            msg = (f'{obj.func.name}() takes {len(obj.func.args)} parameters '
                   f'but {len(obj.params)} were given\n'
                   f'loc: {repr(get_location(obj))}')
            raise TypeError(msg)
    else:
        msg = 'call function must be FunctionDefinition or ObjectImport'
        raise TypeError(msg)


def check_functions_processor(model, metamodel):
    """processor to check function definitions and function calls"""
    for func in get_children_of_type('FunctionDefinition', model):
        check_function_definition(func, metamodel)
    for call in get_children_of_type('FunctionCall', model):
        check_function_call(call, metamodel)
