# pylint: disable=protected-access
"""
type checker based on a simple recursion
"""
from textx import get_location
from textx import get_children_of_type, get_parent_of_type
from textx import get_metamodel
from textx import textx_isinstance
from imports import get_object_import


comparisons = ['Equal', 'LessThan', 'GreaterThan', 'LessThanOrEqual',
               'GreaterThanOrEqual']


class ExpressionTypeError(TypeError):
    """raise this exception if an invalid type is used in an expression"""
    def __init__(self, obj):
        self.obj = obj
        name = obj.__class__.__name__
        message = 'invalid type in '+name+' at '+repr(get_location(obj))
        super().__init__(message)


class TypeMismatchError(TypeError):
    """raise this exception if two objects have incompatible types"""
    def __init__(self, obj1, obj2):
        message = ('Type mismatch: \n' +
                   repr(obj1.type_) + repr(get_location(obj1)) + '\n' +
                   repr(obj2.type_) + repr(get_location(obj2)))
        super().__init__(message)


def check_types_processor(model, metamodel):
    """evaluate type of all objects that have type"""
    classes = ['VarTuple', 'Tuple', 'Variable', 'Expression', 'Or',
               'Equal', 'LessThan', 'GreaterThan', 'LessThanOrEqual',
               'GreaterThanOrEqual', 'IfExpression', 'IfFunction',
               'FunctionCall']
    for cls in classes:
        for obj in get_children_of_type(cls, model):
            if not get_parent_of_type('FunctionDefinition', obj):
                parent_call = get_parent_of_type('FunctionCall', obj)
                if parent_call is not None:
                    _ = parent_call.type_
                _ = obj.type_
                # print(obj, obj.type_)  # replace with a debug-level log


def variable_type(self):
    """evaluate variable type"""
    if not hasattr(self, '__type'):
        if textx_isinstance(self.parent, get_metamodel(self)['VarTuple']):
            vars_ = self.parent.variables
            types = self.parent.parameter.type_
            self.__type = next(t for var, t in zip(vars_, types) if var == self)
        else:
            self.__type = self.parameter.type_
    return self.__type


def expression_type(self):
    """evaluate expression type"""
    # a possible problem to distiguish (int, float) from tuple or table type
    if not hasattr(self, '__type'):
        if all(op.type_ == float for op in self.operands):
            self.__type = float
        elif all(op.type_ == int for op in self.operands):
            self.__type = int
        elif all(op.type_ in (int, float) for op in self.operands):
            self.__type = float
        elif all(op.type_ in (int, (int, float)) for op in self.operands):
            self.__type = (int, float)
        elif all(op.type_ in (int, float, (int, float)) for op in self.operands):
            self.__type = float
        else:
            raise ExpressionTypeError(self)
    return self.__type


def term_type(self):
    """evaluate term type"""
    # a possible problem to distiguish (int, float) from tuple or table type
    if not hasattr(self, '__type'):
        otypes = [o.type_ for o in self.operands]
        if all(t == int for t in otypes) and all(o != '/' for o in self.operators):
            self.__type = int
        elif all(t in (int, float) for t in otypes):
            self.__type = float
        elif all(t in (int, (int, float)) for t in otypes):
            self.__type = (int, float)
        elif all(t in (int, float, (int, float)) for t in otypes):
            self.__type = float
        else:
            raise ExpressionTypeError(self)
    return self.__type


def factor_type(self):
    """evaluate factor type"""
    if not hasattr(self, '__type'):
        self.__type = self.operand.type_
    return self.__type


def operand_type(self):
    """evaluate operand type"""
    if not hasattr(self, '__type'):
        if self.op_val is not None:
            self.__type = type(self.op_val)
        elif self.op_id:
            self.__type = self.op_id.type_
        elif self.op_func:
            self.__type = self.op_func.type_
        else:
            self.__type = self.op_expr.type_
    return self.__type


def boolean_expression_type(self):
    """evaluate boolean expression type"""
    try:
        if hasattr(self, 'operand'):
            assert self.operand.type_ == bool
        else:
            assert all(o.type_ == bool for o in self.operands)
    except AssertionError as err:
        raise ExpressionTypeError(self) from err
    return bool


def if_expression_type(self):
    """evaluate if-expression type"""
    try:
        assert self.expr.type_ == bool
    except AssertionError as err:
        raise ExpressionTypeError(self) from err
    try:
        assert self.true_.type_ == self.false_.type_
    except AssertionError as err:
        raise TypeMismatchError(self.true_, self.false_) from err
    return self.true_.type_


def comparison_type(self):
    """evaluate comparison type"""
    # a possible problem to distiguish (int, float) from tuple or table type
    num_type = (int, float, (int, float))
    numbers = self.left.type_ in num_type and self.right.type_ in num_type
    booleans = self.left.type_ == bool and self.right.type_ == bool
    strings = self.left.type_ == str and self.right.type_ == str
    unbound = self.left.type_ is None and self.right.type_ is None
    try:
        assert numbers or booleans or strings or unbound
    except AssertionError as err:
        raise TypeMismatchError(self.left, self.right) from err
    return bool


def var_reference_type(self):
    """evaluate the type of a referenced variable"""
    if not hasattr(self, '__type'):
        self.__type = self.ref.type_
    return self.__type


def object_import_type(self):
    """evaluate the type of an imported object"""
    if not hasattr(self, '__type'):
        obj = get_object_import(self)
        assert not callable(obj)
        self.__type = type(obj)
    return self.__type


def function_call_type(self):
    """evaluate the type of a function call"""
    if not hasattr(self, '__type'):
        metamodel = get_metamodel(self)
        if textx_isinstance(self.func, metamodel['ObjectImport']):
            if textx_isinstance(self.parent, metamodel['Operand']):
                self.__type = (int, float)  # isinstance(obj, __type) works
            elif textx_isinstance(self.parent, metamodel['BooleanOperand']):
                self.__type = bool
            elif any(textx_isinstance(self.parent, metamodel[c]) for c in comparisons):
                other = self.parent.right if self.parent.left is self else self.parent.left
                if not ((textx_isinstance(other, metamodel['FunctionCall']) and
                         textx_isinstance(other.func, metamodel['ObjectImport'])) or
                        (textx_isinstance(other, metamodel['VarReference']) and
                         textx_isinstance(other.ref, metamodel['ObjectImport']))):
                    if other.type_ in (int, float):
                        self.__type = (int, float)
                    else:
                        self.__type = other.type_
                else:
                    self.__type = None
            elif (textx_isinstance(self.parent, metamodel['IfFunction']) or
                  textx_isinstance(self.parent, metamodel['IfExpression'])):
                other = self.parent.true_ if self.parent.false_ is self else self.parent.false_
                if not ((textx_isinstance(other, metamodel['FunctionCall']) and
                         textx_isinstance(other.func, metamodel['ObjectImport'])) or
                        (textx_isinstance(other, metamodel['VarReference']) and
                         textx_isinstance(other.ref, metamodel['ObjectImport']))):
                    self.__type = other.type_
                else:
                    self.__type = None
            elif textx_isinstance(self.parent, metamodel['VarTuple']):
                self.__type = tuple([None]*len(self.parent.variables))
            else:
                self.__type = None
        elif textx_isinstance(self.func, metamodel['FunctionDefinition']):
            if not get_parent_of_type('FunctionDefinition', self):
                try:
                    self.__type = self.expr.type_
                except ExpressionTypeError as err:
                    my_name = self.__class__.__name__
                    el_name = err.obj.__class__.__name__
                    msgs = ['invalid type in', el_name, 'at',
                            repr(get_location(err.obj)), 'from', my_name, 'at',
                            repr(get_location(self))]
                    raise TypeError(' '.join(msgs)) from err
            else:
                self.__type = None  # the expression is replicated in calls
    return self.__type


def tuple_type(self):
    """evaluate the types in a tuple object"""
    if not hasattr(self, '__type'):
        self.__type = tuple(param.type_ for param in self.params)
    return self.__type


def var_tuple_type(self):
    """evaluate / check the types in a variables tuple object"""
    if not hasattr(self, '__type'):
        self.__type = self.parameter.type_
        try:
            assert isinstance(self.__type, tuple)
            assert len(self.variables) == len(self.__type)
        except AssertionError as err:
            msg = f'expected {len(self.variables)} variable parameters'
            if isinstance(self.__type, tuple):
                msg += f' but {len(self.__type)} were given'
            msg += f'\nloc: {repr(get_location(self.parameter))}'
            raise ValueError(msg) from err
    return self.__type


def series_type(self):
    """evaluate / check the type of a series"""
    if not hasattr(self, '__type'):
        if not self.need_input:
            types = set(element.type_ for element in self.elements)
            types.discard(None)
            try:
                assert len(types) == 1
            except AssertionError as err:
                msg = f'series elements must have one type but {len(types)}'
                msg += f' types were found:\nloc: {repr(get_location(self))}'
                raise TypeError(msg) from err
            else:
                self.__type = (next(iter(types)), len(self.elements))
        else:
            self.__type = None
    return self.__type


def table_type(self):
    """evaluate / check the types in a table object"""
    if not hasattr(self, '__type'):
        if not self.need_input:
            meta = get_metamodel(self)
            ctype = meta['Series']
            if self.columns_tuple:
                msg = ('a table column must be series or a reference:\n'
                       f'loc: {repr(get_location(self))}')
                for par in self.columns_tuple.params:
                    if textx_isinstance(par, meta['Series']):
                        self.columns.append(par)
                    elif textx_isinstance(par, meta['VarReference']):
                        try:
                            assert textx_isinstance(par.ref.parameter, ctype)
                        except AssertionError as err:
                            raise TypeError(msg) from err
                        self.columns.append(par.ref.parameter)
                    else:
                        raise TypeError(msg)
            assert all(textx_isinstance(c, ctype) for c in self.columns)
            types = [c.type_ for c in self.columns]
            sizes = set(map(lambda x: x[1], types))
            try:
                assert len(sizes) == 1
            except AssertionError as err:
                msg = f'table columns must have one size but {len(sizes)} sizes '
                msg += f'were found:\nloc: {repr(get_location(self))}'
                raise ValueError(msg) from err
            else:
                self.__type = (tuple(map(lambda x: x[0], types)), next(iter(sizes)))
            names = [c.name for c in self.columns]
            try:
                assert len(set(names)) == len(names)
            except AssertionError as err:
                msg = ('repeating column names were found in table:\n'
                       f'loc: {repr(get_location(self))}')
                raise ValueError(msg) from err
        else:
            self.__type = None
    return self.__type


def get_sliced_type(obj):
    """return a type slice of an iterable/sequence data structure object"""
    type_ = obj.ref.type_[0]
    size_ = obj.ref.type_[1]
    if obj.slice:
        obj_slice = slice(obj.start, obj.stop, obj.step)
        slice_size = len(range(*obj_slice.indices(size_)))
        ret = (type_, slice_size)
    elif obj.start is not None:
        try:
            assert not obj.array
        except AssertionError as err:
            msg = ('array property not valid for plain or tuple types:\n'
                   f'loc: {repr(get_location(obj))}')
            raise ValueError(msg) from err
        try:
            assert abs(obj.start) < size_
        except AssertionError as err:
            raise IndexError('tuple index out of range') from err
        ret = type_
    else:
        ret = (type_, size_)
    return ret


def series_property_type(self):
    """evaluate / check the type in a series property object"""
    if not hasattr(self, '__type'):
        assert textx_isinstance(self.ref, get_metamodel(self)['Series'])
        if self.name_:
            self.__type = str
        else:
            self.__type = get_sliced_type(self)
    return self.__type


def object_property_type(self):
    """evaluate / check the type in a object property object"""
    if not hasattr(self, '__type'):
        if textx_isinstance(self.ref, get_metamodel(self)['Table']):
            try:
                assert not self.name_
            except AssertionError as err:
                msg = ('Table object has no property "name":\n'
                       f'loc: {repr(get_location(self))}')
                raise AttributeError(msg) from err
            if self.columns:
                self.__type = (str,)*len(self.ref.columns)
            else:
                self.__type = get_sliced_type(self)
        elif textx_isinstance(self.ref, get_metamodel(self)['Series']):
            try:
                assert not self.columns
            except AssertionError as err:
                msg = ('Series object has no property "columns":\n'
                       f'loc: {repr(get_location(self))}')
                raise AttributeError(msg) from err
            self.__type = str if self.name_ else get_sliced_type(self)
        else:
            msg = 'only table and series are currently supported'
            raise NotImplementedError(msg)
    return self.__type


def add_type_properties(metamodel):
    """Add object class properties using monkey style patching"""
    metamodel['Variable'].type_ = property(variable_type)
    metamodel['VarReference'].type_ = property(var_reference_type)
    metamodel['Factor'].type_ = property(factor_type)
    metamodel['Term'].type_ = property(term_type)
    metamodel['Expression'].type_ = property(expression_type)
    metamodel['Operand'].type_ = property(operand_type)
    # a possible problem to distiguish (int, float) from tuple or table type
    metamodel['Number'].type_ = (int, float)
    metamodel['Bool'].type_ = bool
    metamodel['String'].type_ = str
    metamodel['Null'].type_ = type(None)
    metamodel['BooleanOperand'].type_ = property(operand_type)
    metamodel['And'].type_ = property(boolean_expression_type)
    metamodel['Or'].type_ = property(boolean_expression_type)
    metamodel['Not'].type_ = property(boolean_expression_type)
    metamodel['IfExpression'].type_ = property(if_expression_type)
    metamodel['IfFunction'].type_ = property(if_expression_type)
    metamodel['Equal'].type_ = property(comparison_type)
    metamodel['GreaterThan'].type_ = property(comparison_type)
    metamodel['LessThan'].type_ = property(comparison_type)
    metamodel['LessThanOrEqual'].type_ = property(comparison_type)
    metamodel['GreaterThanOrEqual'].type_ = property(comparison_type)
    metamodel['ObjectImport'].type_ = property(object_import_type)
    metamodel['FunctionCall'].type_ = property(function_call_type)
    metamodel['Tuple'].type_ = property(tuple_type)
    metamodel['VarTuple'].type_ = property(var_tuple_type)
    metamodel['Series'].type_ = property(series_type)
    metamodel['Table'].type_ = property(table_type)
    metamodel['SeriesProperty'].type_ = property(series_property_type)
    metamodel['ObjectProperty'].type_ = property(object_property_type)
