# pylint: disable=protected-access
"""
Language interpreter for immediate local evaluation using python
"""
import pandas
from textx import textx_isinstance
from textx import get_metamodel
from textx import get_children_of_type, get_parent_of_type
from constraints import get_object_import
import ioops


def is_active_branch(obj):
    """obj is either a statement, a function or an expression"""
    metamodel = get_metamodel(obj)
    class_names = ['IfStatement', 'IfFunction', 'IfExpression']
    mm_classes = [metamodel[cl] for cl in class_names]
    if any(textx_isinstance(obj.parent, cls) for cls in mm_classes):
        par = obj.parent
        ret = (par.true_ if par.expr.value else par.false_) is obj
    else:
        ret = True
    return ret


def program_value(self):
    """
    Evaluate variables regardless of the order of definition and returns only
    printed values in a single string. The output objects are evaluated in order
    of definition. The output objects are currently processed in this order:
    (ObjectTo, Print)
    """
    for obj in get_children_of_type('ObjectTo', self):
        _ = obj.value
    vals = [p.value for p in get_children_of_type('Print', self) if p.value]
    return '\n'.join(vals)


def variable_value(self):
    """Evaluate a variable"""
    if not hasattr(self, '__value'):
        if textx_isinstance(self.parent, get_metamodel(self)['VarTuple']):
            vars_ = self.parent.variables
            vals_ = self.parent.value
            self.__value = next(val for var, val in zip(vars_, vals_) if var == self)
        else:
            self.__value = self.parameter.value if is_active_branch(self) else None
    return self.__value


def expression_value(self):
    """Evaluate an arithmetic expression"""
    if not hasattr(self, '__value'):
        if is_active_branch(self):
            self.__value = self.operands[0].value
            for operation, operand in zip(self.operators, self.operands[1:]):
                if operation == '+':
                    self.__value += operand.value
                else:
                    self.__value -= operand.value
        else:
            self.__value = None
    return self.__value


def term_value(self):
    """Evaluate a term"""
    if not hasattr(self, '__value'):
        self.__value = self.operands[0].value
        for operation, operand in zip(self.operators, self.operands[1:]):
            if operation == '*':
                self.__value *= operand.value
            else:
                self.__value /= operand.value
    return self.__value


def factor_value(self):
    """Evaluate a factor"""
    if not hasattr(self, '__value'):
        value = self.operand.value
        self.__value = -value if self.sign == '-' else value
    return self.__value


def operand_value(self):
    """Evaluate an operand"""
    if not hasattr(self, '__value'):
        if self.op_val is not None:
            self.__value = self.op_val
        elif self.op_id:
            self.__value = self.op_id.value
        elif self.op_func:
            self.__value = self.op_func.value
        else:
            self.__value = self.op_expr.value
    return self.__value


def or_value(self):
    """Evaluate OR expression"""
    if not hasattr(self, '__value'):
        self.__value = self.operands[0].value
        for operand in self.operands[1:]:
            self.__value = self.__value or operand.value
    return self.__value


def and_value(self):
    """Evaluate AND expression"""
    if not hasattr(self, '__value'):
        self.__value = self.operands[0].value
        for operand in self.operands[1:]:
            self.__value = self.__value and operand.value
    return self.__value


def not_value(self):
    """Evaluate NOT expression"""
    if not hasattr(self, '__value'):
        self.__value = not self.operand.value if self.not_ else self.operand.value
    return self.__value


def print_value(self):
    """Evaluate the print function"""
    if not hasattr(self, '__value'):
        if is_active_branch(self):
            self.__value = ' '.join([repr(p.value) for p in self.params])
        else:
            self.__value = None
    return self.__value


def if_statement_value(self):
    """Evaluate an if-statement object, returns a statement object"""
    if not hasattr(self, '__value'):
        if is_active_branch(self):
            self.__value = self.true_ if self.expr.value else self.false_
        else:
            self.__value = None
    return self.__value


def if_expression_value(self):
    """Evaluate an if-expression object, returns the expression value"""
    if not hasattr(self, '__value'):
        self.__value = self.true_.value if self.expr.value else self.false_.value
    return self.__value


def equal_value(self):
    """Evaluate an equal expression object"""
    if not hasattr(self, '__value'):
        self.__value = self.left.value == self.right.value
    return self.__value


def less_than_value(self):
    """Evaluate a less_than expression object"""
    if not hasattr(self, '__value'):
        self.__value = self.left.value < self.right.value
    return self.__value


def greater_than_value(self):
    """Evaluate a greater_than expression object"""
    if not hasattr(self, '__value'):
        self.__value = self.left.value > self.right.value
    return self.__value


def less_than_or_equal_value(self):
    """Evaluate a less_than_or_equal expression object"""
    if not hasattr(self, '__value'):
        self.__value = self.left.value <= self.right.value
    return self.__value


def greater_than_or_equal_value(self):
    """Evaluate a greater_than_or_equal equal expression object"""
    if not hasattr(self, '__value'):
        self.__value = self.left.value >= self.right.value
    return self.__value


def var_reference_value(self):
    """Evaluate a VarReference object"""
    if not hasattr(self, '__value'):
        self.__value = self.ref.value
    return self.__value


def object_import_value(self):
    """Evaluate an imported non-callable object"""
    if not hasattr(self, '__value'):
        obj = get_object_import(self)
        assert not callable(obj)
        self.__value = obj
    return self.__value


def function_call_value(self):
    """Evaluate a function_call object"""
    if not hasattr(self, '__value'):
        metamodel = get_metamodel(self)
        if textx_isinstance(self.func, metamodel['ObjectImport']):
            obj = get_object_import(self.func)
            assert callable(obj)
            par_values = [p.value for p in self.params]
            self.__value = obj(*par_values)
        elif textx_isinstance(self.func, metamodel['FunctionDefinition']):
            if not get_parent_of_type('FunctionDefinition', self):
                self.__value = self.expr.value
            else:
                self.__value = None
    return self.__value


def tuple_value(self):
    """Evaluate a tuple object"""
    if not hasattr(self, '__value'):
        self.__value = tuple(param.value for param in self.params)
    return self.__value


def series_value(self):
    """Evaluate a series object"""
    if not hasattr(self, '__value'):
        if self.need_input:
            self.load()
        else:
            elements = (e.value for e in self.elements)
            self.__value = pandas.Series(name=self.name, data=elements)
    return self.__value


def table_value(self):
    """Evaluate a table object"""
    if not hasattr(self, '__value'):
        if self.need_input:
            self.load()
        else:
            self.__value = pandas.concat((c.value for c in self.columns), axis=1)
    return self.__value


def get_sliced_value(obj):
    """return a value slice of an iterable/sequence data structure object"""
    value = obj.ref.value
    if obj.slice:
        ret = value[obj.start:obj.stop:obj.step]
    elif obj.start is not None:
        if hasattr(obj.ref, 'columns'):  # table
            dfr = value[obj.start:obj.start+1]
            ret = list(dfr.itertuples(index=False, name=None))[0]  # tuple
        else:  # series
            ret = value[obj.start]  # plain type
    else:
        ret = value
    return ret.to_numpy() if obj.array else ret


def object_property_value(self):
    """Evaluate an object property"""
    if not hasattr(self, '__value'):
        if hasattr(self, 'name_') and self.name_:
            self.__value = self.ref.name
        elif hasattr(self, 'columns') and self.columns:
            self.__value = self.ref.value.columns.to_list()
        else:
            self.__value = get_sliced_value(self)
    return self.__value


def plain_type_value(self):
    """Evaluate an object of plain type"""
    if self.__value is None:
        if self.need_input:
            self.load()
    return self.__value


def object_to_value(self):
    """Store an object to file or url"""
    if not hasattr(self, '__value'):
        obj_val = self.ref.parameter.value
        if isinstance(obj_val, pandas.Series):
            val = dict(data=obj_val.to_list(), name=obj_val.name)
        elif isinstance(obj_val, pandas.DataFrame):
            val = obj_val.to_dict(orient='list')
        else:
            val = obj_val
        try:
            ioops.store_value(val, self.url, self.filename)
        except OSError:
            self.__value = False
            raise
        else:
            self.__value = True
    return self.__value


def add_value_properties(metamodel):
    """Add object class properties using monkey style patching"""
    metamodel['Program'].value = property(program_value)
    metamodel['Variable'].value = property(variable_value)
    metamodel['VarReference'].value = property(var_reference_value)
    metamodel['Factor'].value = property(factor_value)
    metamodel['Term'].value = property(term_value)
    metamodel['Expression'].value = property(expression_value)
    metamodel['Operand'].value = property(operand_value)
    metamodel['BooleanOperand'].value = property(operand_value)
    metamodel['And'].value = property(and_value)
    metamodel['Or'].value = property(or_value)
    metamodel['Not'].value = property(not_value)
    metamodel['Print'].value = property(print_value)
    metamodel['ObjectTo'].value = property(object_to_value)
    metamodel['IfStatement'].value = property(if_statement_value)
    metamodel['IfExpression'].value = property(if_expression_value)
    metamodel['IfFunction'].value = property(if_expression_value)
    metamodel['Equal'].value = property(equal_value)
    metamodel['LessThan'].value = property(less_than_value)
    metamodel['GreaterThan'].value = property(greater_than_value)
    metamodel['LessThanOrEqual'].value = property(less_than_or_equal_value)
    metamodel['GreaterThanOrEqual'].value = property(greater_than_or_equal_value)
    metamodel['ObjectImport'].value = property(object_import_value)
    metamodel['FunctionCall'].value = property(function_call_value)
    metamodel['Tuple'].value = property(tuple_value)
    metamodel['VarTuple'].value = property(variable_value)
    metamodel['Series'].value = property(series_value)
    metamodel['Table'].value = property(table_value)
    metamodel['SeriesProperty'].value = property(object_property_value)
    metamodel['ObjectProperty'].value = property(object_property_value)
    metamodel['Null'].value = None
    metamodel['String'].value = property(plain_type_value)
    metamodel['Number'].value = property(plain_type_value)
    metamodel['Bool'].value = property(plain_type_value)
