"""
tests for basic language components
"""
import os
import pytest
from textx import metamodel_from_file
from textx import get_children_of_type
from textx.exceptions import TextXSyntaxError
from textx.exceptions import TextXSemanticError
from constraints import InitializationError
from constraints import add_processors
from metamodel import add_properties


@pytest.fixture(name='metamodel')
def fixture_metamodel():
    """parse the grammar and generate the object classes"""
    this_folder = os.path.abspath(os.path.dirname(__file__))
    mm_file = os.path.join(this_folder, '../grammar/virtmat.tx')
    mm_obj = metamodel_from_file(mm_file, auto_init_attributes=False)
    add_properties(mm_obj)
    add_processors(mm_obj)
    return mm_obj


def test_empty_program(metamodel):
    """test empty programs"""
    prog_inps = ['', ' ', '\n', '# comment', ' # \n#']
    for prog_inp in prog_inps:
        prog = metamodel.model_from_str(prog_inp)
        assert isinstance(prog, str)
        assert prog == ''


def test_multiline_comments(metamodel):
    """test python-like multiline comment"""
    prog_str = ('"""\nThis is a test set for the scientific computing language'
                '\n"""\n\n""" second multiline comment\n """\n\n"""third multil'
                'ine comment"""')
    prog = metamodel.model_from_str(prog_str)
    assert isinstance(prog, str)
    assert prog == ''


def test_invalid_statements(metamodel):
    """test expressions and empty statements in program"""
    prog_inps = ["a == 1", "b", ";", "c = 1; c + 1"]
    for inp in prog_inps:
        with pytest.raises(TextXSyntaxError):
            metamodel.model_from_str(inp)


def test_print(metamodel):
    """test print builtin function"""
    prog_str = ("print('8', 8)\nprint('None', null)")
    prog = metamodel.model_from_str(prog_str)
    objs = get_children_of_type('Print', prog)
    assert len(objs) == 2
    assert prog.value == "'8' 8\n'None' None"


def test_string_literal(metamodel):
    """test variable initialization with a string literal"""
    prog_str = ("string = 'Abc'")
    prog = metamodel.model_from_str(prog_str)
    var_objs = get_children_of_type('Variable', prog)
    var_strs = get_children_of_type('String', prog)
    assert isinstance(prog.value, str)
    assert prog.value == ''
    assert len(var_objs) == 1
    assert len(var_strs) == 1
    assert var_objs[0].type_ == str
    assert var_objs[0].value == 'Abc'
    assert var_strs[0].type_ == str
    assert var_strs[0].value == 'Abc'


def test_int_literal(metamodel):
    """test variable initialization with an integer literal"""
    prog_str = ('a = 10')
    prog = metamodel.model_from_str(prog_str)
    var_objs = get_children_of_type('Variable', prog)
    exp_objs = get_children_of_type('Expression', prog)
    assert isinstance(prog.value, str)
    assert prog.value == ''
    assert len(var_objs) == 1
    assert len(exp_objs) == 1
    assert var_objs[0].type_ == int
    assert var_objs[0].value == 10
    assert exp_objs[0].type_ == int
    assert exp_objs[0].value == 10


def test_int_expression(metamodel):
    """test variable initialization with an int arithmetic expression"""
    prog_str = ('a = 10; b = 2 * a + 17')
    prog = metamodel.model_from_str(prog_str)
    var_objs = get_children_of_type('Variable', prog)
    exp_objs = get_children_of_type('Expression', prog)
    assert isinstance(prog.value, str)
    assert prog.value == ''
    assert len(var_objs) == 2
    assert len(exp_objs) == 2
    for var in var_objs:
        if var.name == 'b':
            assert var.type_ == int
            assert var.value == 37
        if var.name == 'a':
            assert var.type_ == int
            assert var.value == 10


def test_float_expression(metamodel):
    """test variable initialization with an float arithmetic expression"""
    prog_str = ('a = 10; b = 2*a + 17 \nc = -(4-1)*a + (2+4.67) +b*5.89/(.2+7)')
    prog = metamodel.model_from_str(prog_str)
    var_objs = get_children_of_type('Variable', prog)
    assert prog.value == ''
    assert len(var_objs) == 3
    for var in var_objs:
        if var.name == 'c':
            assert var.type_ == float
            assert var.value == pytest.approx(6.93805555)


def test_undefined_variable_reference(metamodel):
    """reference to an undefined variable"""
    prog_str = ('a = b')
    with pytest.raises(TextXSemanticError,
                       match='Unknown object "b" of class "GeneralVariable"'):
        metamodel.model_from_str(prog_str)


def test_repeated_initialization_reference(metamodel):
    """repeated initialization of a variable that has been referenced"""
    prog_str = ('a = 10; b = 2*a + 17; a = 5')
    with pytest.raises(InitializationError, match='repeated initialization'):
        metamodel.model_from_str(prog_str)


def test_repeated_initialization_nonreference(metamodel):
    """repeated initialization of a variable that has not been referenced"""
    prog_str = ('a = 10; b = 2*a + 17; b = 5')
    with pytest.raises(InitializationError, match='obj: b'):
        metamodel.model_from_str(prog_str)


def test_boolean_literal(metamodel):
    """test variable initialization with boolean literals"""
    prog_str = ('foo = true; baz = false; bar = not (true and false) and not true; '
                'foobar = true or false')
    prog = metamodel.model_from_str(prog_str)
    var_objs = get_children_of_type('Variable', prog)
    assert prog.value == ''
    assert len(var_objs) == 4
    for var in var_objs:
        if var.name == 'foo':
            assert var.type_ == bool
            assert var.value is True
        if var.name == 'bar':
            assert var.type_ == bool
            assert var.value is False
        if var.name == 'foobar':
            assert var.type_ == bool
            assert var.value is True


def test_boolean_expression(metamodel):
    """test variable initialization with boolean expressions"""
    prog_str = ('blah = not (true and false) and not true; foo = not blah; '
                'foobar = true or false; bar = (foo and blah) or foobar')
    prog = metamodel.model_from_str(prog_str)
    var_objs = get_children_of_type('Variable', prog)
    assert prog.value == ''
    assert len(var_objs) == 4
    for var in var_objs:
        if var.name == 'bar':
            assert var.type_ == bool
            assert var.value is True
        if var.name == 'foo':
            assert var.type_ == bool
            assert var.value is True
        if var.name == 'blah':
            assert var.type_ == bool
            assert var.value is False
        if var.name == 'foobar':
            assert var.type_ == bool
            assert var.value is True


def test_illegal_use_of_keywords(metamodel):
    """test illegal use of keywords"""
    prog_strs = ['not = 0', 'true = false', 'all = 4', 'select = 5', 'null = 4']
    for prog_str in prog_strs:
        with pytest.raises(TextXSyntaxError):
            metamodel.model_from_str(prog_str)


def test_substring_keyword(metamodel):
    """test variable names that have keywords as substrings"""
    prog_strs = ['select_number = 5', 'notbar= 3.14', 'true_ = 1', '_orkan = 3',
                 "andy = 'Hi!'", 'truefalse = true or false', 'nullvar = null']
    for prog_str in prog_strs:
        metamodel.model_from_str(prog_str)


def test_noeol_after_keyword(metamodel):
    """test a keyword at end of string with no eol"""
    prog_str = 'a = true'  # keyword doesn't get parsed without eol character
    with pytest.raises(TextXSemanticError,
                       match='Unknown object "true" of class "GeneralVariable"'):
        metamodel.model_from_str(prog_str)


def test_eol_after_keyword(metamodel):
    """test a keyword at end of string followed by eol character"""
    prog_strs = ['a = true\n']  # keyword gets parsed correctly with eol
    for prog_str in prog_strs:
        metamodel.model_from_str(prog_str)


def test_space_after_eol(metamodel):
    """a statement cannot be parsed if there is white space from eol"""
    prog_strs = ['b = 4\n a = 2']
    for prog_str in prog_strs:
        with pytest.raises(TextXSyntaxError):
            metamodel.model_from_str(prog_str)


def test_semicolon(metamodel):
    """test missing statement before/after semicolon"""
    prog_strs = ['; b = 1', 'a = 3;']
    for prog_str in prog_strs:
        with pytest.raises(TextXSyntaxError):
            metamodel.model_from_str(prog_str)


def test_unordered_boolean(metamodel):
    """test unordered initialization - boolean"""
    prog_str = "bar = true; foobar = (foo or bar); foo = not bar"
    prog = metamodel.model_from_str(prog_str)
    var_list = get_children_of_type('Variable', prog)
    var = next(v for v in var_list if v.name == 'foobar')
    assert var.type_ == bool
    assert var.value is True


def test_unordered_float(metamodel):
    """test unordered initialization - float"""
    prog_str = "a = b + c; b = c + 2.; c = 0"
    prog = metamodel.model_from_str(prog_str)
    var_list = get_children_of_type('Variable', prog)
    var = next(v for v in var_list if v.name == 'a')
    assert var.type_ == float
    assert var.value == 2.0


def test_if_statement_with_prints(metamodel):
    """test if-statement containing print statements"""
    prog_inps = ['bar = true; if bar then print(\'true branch\') else print(\'false branch\')',
                 'bar = true; if not bar then print(true) else print(false)']
    prog_outs = ['\'true branch\'', 'False']
    for prog_inp, prog_out in zip(prog_inps, prog_outs):
        prog = metamodel.model_from_str(prog_inp)
        print(f'>{prog.value}<')
        assert prog.value == prog_out


def test_if_statement_with_assignments(metamodel):
    """test if-statement containing variable initializations"""
    prog_inp = "c = 3; a = 10; if false then e = 10*c else f = 5*a"
    prog = metamodel.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    var = next(v for v in var_list if v.name == 'f')
    assert var.type_ == int
    assert var.value == 50
    var = next(v for v in var_list if v.name == 'e')
    assert var.type_ == int
    assert var.value is None


def test_if_statement_with_assignment_global_scope(metamodel):
    """test if-statement with variable initializations in global scope"""
    inp = 'if false then a = 10 else b = 5; c = b\n'
    metamodel.model_from_str(inp)


def test_multiline_if_statement(metamodel):
    """test a multi-line if-statement"""
    prog_inp = ("bexpr = (not (true and false) and not true)\nif (not (true and"
                " false) and not true)\n  then\n    print('bexpr', true, bexpr)"
                "\n  else\n    print('bexpr', false, bexpr)")
    prog = metamodel.model_from_str(prog_inp)
    assert prog.value == '\'bexpr\' False False'


def test_if_expression_active_branch(metamodel):
    """test if-expression in an active branch"""
    prog_inp = ("bar = true; foo = true; foobar = bar if not bar else foo\n"
                "foobar2 = 3 / 1 if not (true and false) and not true else 7.")
    prog = metamodel.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    var = next(v for v in var_list if v.name == 'foobar')
    assert var.type_ == bool
    assert var.value is True
    var = next(v for v in var_list if v.name == 'foobar2')
    assert var.type_ == float
    assert var.value == 7.


def test_if_expression_non_active_branch(metamodel):
    """test if-expression in an non-active branch"""
    prog_inp = "a = 3; b = (a + 2) if false else 1"
    prog = metamodel.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    var = next(v for v in var_list if v.name == 'b')
    assert var.type_ == int
    assert var.value == 1
    expr_list = get_children_of_type('Expression', prog)
    assert len(expr_list) == 4
    assert expr_list[1].type_ == int
    assert expr_list[1].value is None


def test_if_function(metamodel):
    """test built-in if-function"""
    prog_inp = ("bar = true; true_branch = 'true_branch'\n"
                "p = if(not bar, true_branch, 'false_branch')")
    prog = metamodel.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    var = next(v for v in var_list if v.name == 'p')
    assert var.type_ == str
    assert var.value == 'false_branch'
    var = next(v for v in var_list if v.name == 'true_branch')
    assert var.type_ == str
    assert var.value == 'true_branch'


def test_if_function_float_expressions(metamodel):
    """test if-function with a float expression in a non-active branch"""
    prog_inp = "bar = true; a = 10; b = a*2+17; c = if(bar, (b+4)/2, 1/2)"
    prog = metamodel.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    var = next(v for v in var_list if v.name == 'c')
    assert var.type_ == float
    assert var.value == 20.5
    expr_list = get_children_of_type('Expression', prog)
    assert len(expr_list) == 5
    assert expr_list[4].type_ == float
    assert expr_list[4].value is None


def test_nested_if_function(metamodel):
    """test nested if-functions"""
    prog_inp = ("bar = true; a = 0; b = 1\n"
                "p = if (bar, if(bar, 2, a), if(false, a, b))")
    prog = metamodel.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    var = next(v for v in var_list if v.name == 'p')
    assert var.type_ == int
    assert var.value == 2


def test_comparison_literal_operands_assignment(metamodel):
    """with arithmetic constant operands in variable assignment"""
    prog_inp = ("a = 3 == 3; b = (4 == 2); c = 4 > 2; d = (-3 < 0)\n"
                "e = (4/2 == 2)")
    prog = metamodel.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    for var in var_list:
        assert var.type_ == bool
    assert next(v for v in var_list if v.name == 'a').value is True
    assert next(v for v in var_list if v.name == 'b').value is False
    assert next(v for v in var_list if v.name == 'c').value is True
    assert next(v for v in var_list if v.name == 'd').value is True
    assert next(v for v in var_list if v.name == 'e').value is True


def test_comparison_literal_operands_if_function(metamodel):
    """with arithmetic constant operands in if-function"""
    prog_inp = ("a = if((-3*3 > 10), true, false);"
                "b = if((-3*3 <= 10), true, false)\n"
                "c = if(3 == 2, -1, +1); d = if(3 >= 2, 1, 0)")
    prog = metamodel.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    assert next(v for v in var_list if v.name == 'a').value is False
    assert next(v for v in var_list if v.name == 'b').value is True
    assert next(v for v in var_list if v.name == 'c').value == 1
    assert next(v for v in var_list if v.name == 'd').value == 1


def test_comparison_expression_operands_assignment(metamodel):
    """comparison with arithmetic expression operands in variable assignment"""
    prog_inp = ("a = 1.0e0; b = 10; c = a == b; d = a > 3*b; e = 10*a-b <= 0;"
                "f = b == b; g = a < 0; h = 2*a-1 == a")
    prog = metamodel.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    assert next(v for v in var_list if v.name == 'a').type_ == float
    assert next(v for v in var_list if v.name == 'b').type_ == int
    assert next(v for v in var_list if v.name == 'c').value is False
    assert next(v for v in var_list if v.name == 'd').value is False
    assert next(v for v in var_list if v.name == 'e').value is True
    assert next(v for v in var_list if v.name == 'f').value is True
    assert next(v for v in var_list if v.name == 'g').value is False
    assert next(v for v in var_list if v.name == 'h').value is True


def test_comparison_expression_operands_boolean_expression(metamodel):
    """comparison with arithmetic expression operands in boolean expressions"""
    prog_inp = ("res = 0\n"
                "ans1 = ((res < 0.0001) and (res >= 0.0)) or\n"
                "((res > -0.0001) and (res <= 0.0))\n"
                "ans2 = not (res == res); ans3 = not (res < res)\n"
                "ans4 = not (res <= res)")
    prog = metamodel.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    assert next(v for v in var_list if v.name == 'ans1').value is True
    assert next(v for v in var_list if v.name == 'ans2').value is False
    assert next(v for v in var_list if v.name == 'ans3').value is True
    assert next(v for v in var_list if v.name == 'ans4').value is False


def test_comparison_boolean_operands(metamodel):
    """comparison with boolean operands"""
    prog_inp = ("a = true == false; b = false == false; c = false == not false;"
                "bar = true; d = bar == not false;"
                "e = not (bar == true and true);"
                "f = not (bar == false) and true")
    prog = metamodel.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    assert next(v for v in var_list if v.name == 'a').value is False
    assert next(v for v in var_list if v.name == 'b').value is True
    assert next(v for v in var_list if v.name == 'c').value is False
    assert next(v for v in var_list if v.name == 'd').value is True
    assert next(v for v in var_list if v.name == 'e').value is False
    assert next(v for v in var_list if v.name == 'f').value is True


def test_comparison_boolean_operands_invalid_operators(metamodel):
    """comparison with boolean operands and invalid comparison operators"""
    prog_inp = ("a = true < false; b = true <= false; c = true > true\n"
                "d = true <= true\n")
    prog = metamodel.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    assert next(v for v in var_list if v.name == 'a').value is False
    assert next(v for v in var_list if v.name == 'b').value is False
    assert next(v for v in var_list if v.name == 'c').value is False
    assert next(v for v in var_list if v.name == 'd').value is True


def test_comparison_string_operands(metamodel):
    """comparison with string operands in assignments and expressions"""
    prog_inp = ("string = 'Abc'; a = string == string; b = string == 'Abc';"
                "c = not (string == 'Q'); d = 'q' == 'q'; e = 'q' == 'Q'")
    prog = metamodel.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    assert next(v for v in var_list if v.name == 'a').value is True
    assert next(v for v in var_list if v.name == 'b').value is True
    assert next(v for v in var_list if v.name == 'c').value is True
    assert next(v for v in var_list if v.name == 'd').value is True
    assert next(v for v in var_list if v.name == 'e').value is False


def test_comparison_if_expression(metamodel):
    """comparison in if-expressions"""
    prog_inp = "a = 3.5 if 3 == 2 else 1.\n"
    prog = metamodel.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    assert next(v for v in var_list if v.name == 'a').value == 1.
