"""
tests cyclic dependencies
"""
import os
import pytest
from textx import metamodel_from_file
from constraints import add_processors
from constraints import CyclicDependencyError


@pytest.fixture(name='metamodel')
def fixture_metamodel():
    """parse the grammar and generate the object classes"""
    this_folder = os.path.abspath(os.path.dirname(__file__))
    mm_file = os.path.join(this_folder, '../grammar/virtmat.tx')
    mm_obj = metamodel_from_file(mm_file, auto_init_attributes=False)
    add_processors(mm_obj)
    return mm_obj


def test_zeroth_order_cyclic(metamodel):
    """test zeroth order cycic dependencies (one-node loop)"""
    inp = 'd = d + 3 \n'
    with pytest.raises(CyclicDependencyError):
        metamodel.model_from_str(inp)


def test_first_order_cyclic(metamodel):
    """test first order cycic dependencies (two-nodes loop)"""
    inp = 'f = 3 * g; g = f + 1 \n'
    with pytest.raises(CyclicDependencyError):
        metamodel.model_from_str(inp)


def test_second_order_cyclic(metamodel):
    """test second order cycic dependencies (three-nodes loop)"""
    inp = 'a = 3 + c; c = 4 * b; b = 1 / a \n'
    with pytest.raises(CyclicDependencyError):
        metamodel.model_from_str(inp)


def test_second_order_cyclic_boolean(metamodel):
    """test second order cycic dependencies with boolean variables"""
    inp = 'a = true and c; c = not b; b = false or a \n'
    with pytest.raises(CyclicDependencyError):
        metamodel.model_from_str(inp)


def test_function_zeroth_order_cyclic(metamodel):
    """test second order cycic dependencies - deepcopy fails"""
    inp = 'cube(x) = x*x*x; a = cube(2+a) \n'
    with pytest.raises(CyclicDependencyError):
        metamodel.model_from_str(inp)


def test_function_first_order_cyclic(metamodel):
    """test first order cycic dependencies - deepcopy fails"""
    inp = 'cube(x) = x*x*x; a = cube(2+b); b = a - 1 \n'
    with pytest.raises(CyclicDependencyError):
        metamodel.model_from_str(inp)


def test_if_cyclic_expr(metamodel):
    """test if function cycic dependencies in expr"""
    inp = 'b = if(c, 3, 4); c = 3 + b > 1 \n'
    with pytest.raises(CyclicDependencyError):
        metamodel.model_from_str(inp)


def test_if_cyclic_in_branch(metamodel):
    """test if function cycic dependencies in branch"""
    inp = 'b = if(false, 3, c); c = 3 + b \n'
    with pytest.raises(CyclicDependencyError):
        metamodel.model_from_str(inp)
