"""
tests for data structures
"""
import os
import pytest
from textx import metamodel_from_file
from textx import get_children_of_type
import metamodel
import constraints


@pytest.fixture(name='meta_model')
def fixture_metamodel():
    """parse the grammar and generate the object classes"""
    this_folder = os.path.abspath(os.path.dirname(__file__))
    mm_file = os.path.join(this_folder, '../grammar/virtmat.tx')
    meta = metamodel_from_file(mm_file, auto_init_attributes=False)
    metamodel.add_properties(meta)
    metamodel.add_processors(meta)
    constraints.add_processors(meta)
    return meta


def test_tuple_assignement_literals(meta_model):
    """test tuple assignment of different literals"""
    inp = '(a, b, c, d) = (1, (2, 3), not false, "Abc")\n'
    prog = meta_model.model_from_str(inp)
    var_list = get_children_of_type('Variable', prog)
    names = ('a', 'b', 'c', 'd')
    values = (1, (2, 3), True, 'Abc')
    types = (int, (int, int), bool, str)
    for ind, name in enumerate(names):
        var = next(v for v in var_list if v.name == name)
        assert var.value == values[ind]
        assert var.type_ == types[ind]


def test_tuple_assignment_variables(meta_model):
    """test assignment to a tuple and tuple assignment of a tuple object"""
    prog = meta_model.model_from_str('a = 1; b = false; c = (a, b); (g, h) = c\n')
    var_list = get_children_of_type('Variable', prog)
    var_c = next(v for v in var_list if v.name == 'c')
    assert var_c.value == (1, False)
    assert var_c.type_ == (int, bool)
    var_g = next(v for v in var_list if v.name == 'g')
    assert var_g.value == 1
    assert var_g.type_ == int
    var_h = next(v for v in var_list if v.name == 'h')
    assert var_h.value is False
    assert var_h.type_ == bool


def test_function_call_returning_tuple(meta_model):
    """test call of a internal function returning a tuple"""
    inp = 'f(x) = (x*x, 1); j = f(2); (k, l) = f(2); (m, n) = (f(2), j)\n'
    prog = meta_model.model_from_str(inp)
    var_list = get_children_of_type('Variable', prog)
    names = ('j', 'k', 'l', 'm', 'n')
    values = ((4, 1), 4, 1, (4, 1), (4, 1))
    types = ((int, int), int, int, (int, int), (int, int))
    for ind, name in enumerate(names):
        var = next(v for v in var_list if v.name == name)
        assert var.value == values[ind]
        assert var.type_ == types[ind]


def test_imported_function_call_returning_tuple(meta_model):
    """test call of imported function returning a tuple"""
    inp = 'use divmod from builtins; (p, r) = divmod(4, 2)\n'
    prog = meta_model.model_from_str(inp)
    var_list = get_children_of_type('Variable', prog)
    var_p = next(v for v in var_list if v.name == 'p')
    assert var_p.value == 2
    assert var_p.type_ is None
    var_r = next(v for v in var_list if v.name == 'r')
    assert var_r.value == 0
    assert var_r.type_ is None


def test_tuple_if_function(meta_model):
    """test a tuple with if function call"""
    inp = '(t, z) = if(false, (1, 0), (0, 1))\n'
    prog = meta_model.model_from_str(inp)
    var_list = get_children_of_type('Variable', prog)
    var_t = next(v for v in var_list if v.name == 't')
    assert var_t.value == 0
    assert var_t.type_ == int
    var_z = next(v for v in var_list if v.name == 'z')
    assert var_z.value == 1
    assert var_z.type_ == int


def test_tuple_assignment_value_error(meta_model):
    """test tuple assignment with parameter mismatch"""
    inp = '(a, b, c) = (1, (2, 3))\n'
    mstr = 'expected 3 variable parameters but 2 were given'
    with pytest.raises(ValueError, match=mstr):
        meta_model.model_from_str(inp)


def test_assignment_of_tuple_with_type_mismatch(meta_model):
    """test assignment of tuple object with type mismatch"""
    inp = 'f(x, y) = (x*x + y*y, 1); e = f(1, true)\n'
    with pytest.raises(TypeError, match='invalid type in Term'):
        meta_model.model_from_str(inp)


def test_tuple_function(meta_model):
    """test function returning a Tuple that contains a VarReference"""
    inp = 'f(x, y) = (x, x*y, y); e = f(2.0, 3.0)\n'
    prog = meta_model.model_from_str(inp)
    var_list = get_children_of_type('Variable', prog)
    var_e = next(v for v in var_list if v.name == 'e')
    assert var_e.value == (2., 6., 3.)
    assert var_e.type_ == (float, float, float)


def test_series_elements_of_different_type(meta_model):
    """series with elements of different type"""
    inp = 's1 = (numbers: 1, 2, true)\n'
    match_str = 'series elements must have one type but 2 types were found'
    with pytest.raises(TypeError, match=match_str):
        meta_model.model_from_str(inp)


def test_table_columns_of_different_size(meta_model):
    """table with columns of different size """
    inp = 't = ((temperature: 0, 1, 3), (pressure: 4., 5.))\n'
    match_str = 'table columns must have one size but 2 sizes were found'
    with pytest.raises(ValueError, match=match_str):
        meta_model.model_from_str(inp)


def test_table_columns_with_repeating_names(meta_model):
    """table with columns with repeating names """
    inp = 't = ((label: 0, 1, 3), (label: 4, 5, 6))\n'
    match_str = 'repeating column names were found in table'
    with pytest.raises(ValueError, match=match_str):
        meta_model.model_from_str(inp)


def test_table_properties(meta_model):
    """test table properties - slice, columns, indexed row, named column"""
    inp = ('t = ((temperature: 0, 1, 3), (pressure: 4., 5., 6.));'
           'a = t:columns; b = t[0]; c = t[0:1]; d = t[::-1];'
           'e = t.temperature; f = t.temperature[0]; g = t.pressure[1:];'
           'h = t.pressure[::-1]:array; i = t.pressure:name\n')
    prog = meta_model.model_from_str(inp)
    var_list = get_children_of_type('Variable', prog)
    var_t = next(v for v in var_list if v.name == 't')
    assert var_t.type_ == ((int, float), 3)
    var_a = next(v for v in var_list if v.name == 'a')
    assert var_a.type_ == (str, str)
    assert var_a.value == ['temperature', 'pressure']
    var_b = next(v for v in var_list if v.name == 'b')
    assert var_b.type_ == (int, float)
    assert var_b.value == (0, 4.0)
    var_c = next(v for v in var_list if v.name == 'c')
    assert var_c.type_ == ((int, float), 1)
    assert var_c.value.to_numpy().tolist()[0] == [0, 4.0]
    var_d = next(v for v in var_list if v.name == 'd')
    assert var_d.type_ == ((int, float), 3)
    assert var_d.value.to_numpy().tolist() == [[3, 6], [1, 5], [0, 4]]
    var_e = next(v for v in var_list if v.name == 'e')
    assert var_e.type_ == (int, 3)
    assert var_e.value.to_numpy().tolist() == [0, 1, 3]
    var_f = next(v for v in var_list if v.name == 'f')
    assert var_f.type_ == int
    assert var_f.value == 0
    var_g = next(v for v in var_list if v.name == 'g')
    assert var_g.type_ == (float, 2)
    assert var_g.value.to_numpy().tolist() == [5., 6.]
    var_h = next(v for v in var_list if v.name == 'h')
    assert var_h.type_ == (float, 3)
    assert var_h.value.tolist() == [6., 5., 4.]
    var_i = next(v for v in var_list if v.name == 'i')
    assert var_i.type_ == str
    assert var_i.value == 'pressure'


def test_series_properties(meta_model):
    """test series properties - name, indexed value, slice, array"""
    inp = ('s1 = (numbers: 1, 2, 3); s2 = (letters: "a", "b", "c");'
           's3 = (booleans: true, false); a = s2; b = s1:name; c = s2[0:1]; '
           'c1 = s3[]; c2 = s3[:1]; c3 = s2[:3:1]; c4 = s2[:]; c5 = s2[::];'
           'd1 = s1[1]; d2 = s1[1:2]; d3 = s1[1:]; e = s2[::-1]:array;'
           'f = s1:array\n')
    prog = meta_model.model_from_str(inp)
    var_list = get_children_of_type('Variable', prog)
    var_s1 = next(v for v in var_list if v.name == 's1')
    assert var_s1.type_ == (int, 3)
    var_b = next(v for v in var_list if v.name == 'b')
    assert var_b.type_ == str
    assert var_b.value == 'numbers'
    var_c = next(v for v in var_list if v.name == 'c')
    assert var_c.type_ == (str, 1)
    assert var_c.value.to_list() == ['a']
    var_d1 = next(v for v in var_list if v.name == 'd1')
    assert var_d1.type_ == int
    assert var_d1.value == 2
    var_d2 = next(v for v in var_list if v.name == 'd2')
    assert var_d2.type_ == (int, 1)
    assert var_d2.value.tolist() == [2]
    var_d3 = next(v for v in var_list if v.name == 'd3')
    assert var_d3.type_ == (int, 2)
    assert var_d3.value.tolist() == [2, 3]
    var_e = next(v for v in var_list if v.name == 'e')
    assert var_e.type_ == (str, 3)
    assert list(var_e.value) == ['c', 'b', 'a']
    var_f = next(v for v in var_list if v.name == 'f')
    assert var_f.type_ == (int, 3)
    assert var_f.value.tolist() == [1, 2, 3]


def test_series_properties_invalid_columns_property(meta_model):
    """test series with invalid columns property"""
    inp = 's = (numbers: 1, 2, 3); g = s:columns\n'
    match_str = 'Series object has no property "columns"'
    with pytest.raises(AttributeError, match=match_str):
        meta_model.model_from_str(inp)


def test_table_properties_invalid_name_property(meta_model):
    """test table with invalid name property"""
    inp = 't = ((temperature: 0, 1, 3)); a = t:name\n'
    match_str = 'Table object has no property "name"'
    with pytest.raises(AttributeError, match=match_str):
        meta_model.model_from_str(inp)


def test_series_properties_invalid_array_property(meta_model):
    """test series with invalid array property"""
    inp = 's = (numbers: 1, 2, 3); g = s[1]:array\n'
    match_str = 'array property not valid for plain or tuple type'
    with pytest.raises(ValueError, match=match_str):
        meta_model.model_from_str(inp)


def test_table_properties_ivalid_array_property(meta_model):
    """test table with invalid array property"""
    inp = 't = ((temperature: 0, 1, 3)); j = t[0]:array\n'
    match_str = 'array property not valid for plain or tuple type'
    with pytest.raises(ValueError, match=match_str):
        meta_model.model_from_str(inp)


def test_object_properties_zero_slice(meta_model):
    """test object slice with an invalid slice step"""
    inp = 's = (numbers: 1, 2, 3); g = s[1:2:0]\n'
    with pytest.raises(ValueError, match='slice step cannot be zero'):
        meta_model.model_from_str(inp)


def test_object_properties_invalid_index(meta_model):
    """test object slice with invalid index"""
    inp = 's = (numbers: 1, 2, 3); g = s[11]\n'
    with pytest.raises(IndexError, match='tuple index out of range'):
        meta_model.model_from_str(inp)


def test_table_with_attribute(meta_model):
    """test table with columns attribute from a tuple"""
    inp = ('s = (booleans: true, false)\n'
           't = Table with columns=(s, (floats: 1.2, -3.5))\n')
    prog = meta_model.model_from_str(inp)
    var_list = get_children_of_type('Variable', prog)
    t_var = next(v for v in var_list if v.name == 't')
    assert t_var.value['floats'][0] == 1.2
    assert t_var.value['booleans'][1].item() is False


def test_table_with_attribute_type_error(meta_model):
    """test table with columns attribute from a tuple a wrong type"""
    inp = 'a = 3; s = (test: 1); t = Table with columns=(a, s)\n'
    match_str = 'a table column must be series'
    with pytest.raises(TypeError, match=match_str):
        meta_model.model_from_str(inp)
