"""
tests for functions
"""
import os
import pytest
from textx import metamodel_from_file
from textx import get_children_of_type
from textx import textx_isinstance
from textx import TextXSemanticError
import metamodel
import constraints
from constraints import InitializationError
from constraints import ArgumentMismatchError


@pytest.fixture(name='meta_model')
def fixture_metamodel():
    """parse the grammar and generate the object classes"""
    this_folder = os.path.abspath(os.path.dirname(__file__))
    mm_file = os.path.join(this_folder, '../grammar/virtmat.tx')
    mm_obj = metamodel_from_file(mm_file, auto_init_attributes=False)
    metamodel.add_properties(mm_obj)
    metamodel.add_processors(mm_obj)
    constraints.add_processors(mm_obj)
    return mm_obj


def test_function_def_empty_argument_list(meta_model):
    """test function definition with empty arguments list"""
    prog_inp = 'func1() = null; func2() = true; func3() = \'Abc\'\n'
    prog = meta_model.model_from_str(prog_inp)
    assert len(get_children_of_type('FunctionDefinition', prog)) == 3


def test_function_def_empty_argument_list_number_var_reference(meta_model):
    """test function definition with empty arguments list"""
    prog_inp = 'a = 2; func1() = a; func2() = 3; func3() = 3 + a\n'
    prog = meta_model.model_from_str(prog_inp)
    assert len(get_children_of_type('FunctionDefinition', prog)) == 3


def test_function_def_unused_arguments(meta_model):
    """test a functin definition with unused arguments"""
    prog_inp = 'f(x, y, z) = 0\n'
    match_str = 'dummy variables do not match function arguments'
    with pytest.raises(ArgumentMismatchError, match=match_str):
        meta_model.model_from_str(prog_inp)


def test_function_def_unknown_arguments(meta_model):
    """test a function definition with unknown"""
    prog_inp = 'g(t) = 2*t + y; f(y) = y\n'
    match_str = 'Unknown object "y" of class "GeneralVariable"'
    with pytest.raises(TextXSemanticError, match=match_str):
        meta_model.model_from_str(prog_inp)


def test_function_def_one_argument(meta_model):
    """test function definition with one argument"""
    prog = meta_model.model_from_str('f(x) = x\n')
    func_defs = get_children_of_type('FunctionDefinition', prog)
    assert len(func_defs) == 1


def test_function_def_two_arithmetic(meta_model):
    """test function definition arithmetic type"""
    prog_inp = 'func(x, y) = a*x + c*y; a = 5+10*c; c = 4\n'
    prog = meta_model.model_from_str(prog_inp)
    func_defs = get_children_of_type('FunctionDefinition', prog)
    assert len(func_defs) == 1


def test_function_def_two_boolean(meta_model):
    """test function definition boolean type"""
    prog_inp = 'func(x, y) = (y or (x and not a)); a = true\n'
    prog = meta_model.model_from_str(prog_inp)
    assert len(get_children_of_type('Or', prog)) == 4
    func_defs = get_children_of_type('FunctionDefinition', prog)
    assert len(func_defs) == 1


def test_function_def_if_expression(meta_model):
    """test function definition with two arguments"""
    prog_inp = 'gt(x, y) = x if x>y else y\n'
    prog = meta_model.model_from_str(prog_inp)
    assert len(get_children_of_type('IfExpression', prog)) == 1
    func_defs = get_children_of_type('FunctionDefinition', prog)
    assert len(func_defs) == 1


def test_function_def_if_function(meta_model):
    """test function definition with two arguments"""
    prog_inp = 'gt(x, y) = if(x>y, x, y)\n'
    prog = meta_model.model_from_str(prog_inp)
    assert len(get_children_of_type('IfFunction', prog)) == 1
    func_defs = get_children_of_type('FunctionDefinition', prog)
    assert len(func_defs) == 1


def test_function_call_empty_argument_list(meta_model):
    """test function call with empty arguments list"""
    prog_inp = 'func() = 1; a = func()\n'
    prog = meta_model.model_from_str(prog_inp)
    funcs = get_children_of_type('FunctionDefinition', prog)
    calls = get_children_of_type('FunctionCall', prog)
    assert len(funcs) == 1
    assert len(calls) == 1
    assert textx_isinstance(calls[0].parent, meta_model['Variable'])
    assert len(calls[0].params) == len(funcs[0].args)
    assert len(calls[0].params) == 0
    assert calls[0].type_ == int
    var_list = get_children_of_type('Variable', prog)
    var = next(v for v in var_list if v.name == 'a')
    assert var.type_ == int
    assert var.value == 1


def test_function_call_one_argument(meta_model):
    """test function call with one argument"""
    prog_inp = 'a = sqr(2); sqr(x) = x*x\n'
    prog = meta_model.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    assert len(var_list) == 1
    funcs = get_children_of_type('FunctionDefinition', prog)
    calls = get_children_of_type('FunctionCall', prog)
    assert len(funcs) == 1
    assert len(calls) == 1
    assert textx_isinstance(calls[0].parent, meta_model['Variable'])
    assert len(calls[0].params) == len(funcs[0].args)
    assert len(calls[0].params) == 1
    assert calls[0].type_ == int
    var = next(v for v in var_list if v.name == 'a')
    assert var.type_ == int
    assert var.value == 4


def test_nested_function_call_one_argument(meta_model):
    """test nested function call, i.e. call with a parameter that is a call"""
    prog_inp = 'a = sqr(sqr(2)); sqr(x) = x*x; b = (sqr((sqr(a)/16))-1)\n'
    prog = meta_model.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    calls = get_children_of_type('FunctionCall', prog)
    assert len(var_list) == 2
    assert len(calls) == 4
    assert all(c.type_ in (int, float) for c in calls)
    var_a = next(v for v in var_list if v.name == 'a')
    var_b = next(v for v in var_list if v.name == 'b')
    assert var_a.type_ == int
    assert var_a.value == 16
    assert var_b.type_ == float
    assert var_b.value == 255.


def test_nested_function_call_two_functions(meta_model):
    """test nested function call with two functions"""
    prog_inp = 'a = diff(sqr(2), sqr(1)); sqr(z) = z*z; diff(x, y) = x - y\n'
    prog = meta_model.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    calls = get_children_of_type('FunctionCall', prog)
    assert len(var_list) == 1
    assert len(calls) == 3
    assert all(c.type_ == int for c in calls)
    var_a = next(v for v in var_list if v.name == 'a')
    assert var_a.type_ == int
    assert var_a.value == 3


def test_function_call_boolean_type(meta_model):
    """test function call of boolean type"""
    prog_inp = 'a = lt(1, 3); lt(x, y) = x < y\n'
    prog = meta_model.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    calls = get_children_of_type('FunctionCall', prog)
    assert len(var_list) == 1
    assert len(calls) == 1
    assert calls[0].type_ == bool
    var_a = next(v for v in var_list if v.name == 'a')
    assert var_a.type_ == bool
    assert var_a.value is True


def test_function_call_if_function(meta_model):
    """test function call with if function"""
    prog_inp = 'a = max(1, 3); max(x, y) = if(x < y, y, x)\n'
    prog = meta_model.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    calls = get_children_of_type('FunctionCall', prog)
    assert len(var_list) == 1
    assert len(calls) == 1
    assert calls[0].type_ == int
    var_a = next(v for v in var_list if v.name == 'a')
    assert var_a.type_ == int
    assert var_a.value == 3


def test_function_definition_with_call(meta_model):
    """test function definition with a call in its expression"""
    prog_inp = 'sqr(x) = x*x; sqr_min_1(y) = (sqr(y)-1); a = sqr_min_1(2)\n'
    prog = meta_model.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    calls = get_children_of_type('FunctionCall', prog)
    assert len(var_list) == 1
    assert len(calls) == 3
    assert all(c.type_ in (int, None) for c in calls)
    var_a = next(v for v in var_list if v.name == 'a')
    assert var_a.type_ == int
    assert var_a.value == 3


def test_function_definition_with_call_boolean(meta_model):
    """test function definition with a call in its expression boolean type"""
    prog_inp = 'sqr(x) = x*x; lt(y, z) = (sqr(y) < z); a = lt(2, 5)\n'
    prog = meta_model.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    calls = get_children_of_type('FunctionCall', prog)
    assert len(var_list) == 1
    assert len(calls) == 3
    assert all(c.type_ in (bool, int, None) for c in calls)
    var_a = next(v for v in var_list if v.name == 'a')
    assert var_a.type_ == bool
    assert var_a.value is True


def test_function_definition_with_call_statement_ordering(meta_model):
    """test function definition with a call with two statement orderings"""
    inp = ('sqr(x) = x*x; one_minus_sqr(y) = sqr(y); a = one_minus_sqr(2);'
           'b = one_minus_sqr2(2); one_minus_sqr2(y) = sqr(y)\n')
    var_list = get_children_of_type('Variable', meta_model.model_from_str(inp))
    assert next(v for v in var_list if v.name == 'a').value == 4
    assert next(v for v in var_list if v.name == 'b').value == 4


def test_function_definition_with_call_dummies_of_different_names(meta_model):
    """test function definition with a call and dummies of different names"""
    inp = ('f1(x) = 2*x; f2(y) = (f1(4) + y); a = f2(2)\n')
    var_list = get_children_of_type('Variable', meta_model.model_from_str(inp))
    assert next(v for v in var_list if v.name == 'a').value == 10


def test_function_definition_with_call_dummies_of_same_names(meta_model):
    """test function definition with a call and dummies of same names"""
    inp = ('f1(x) = 2*x; f2(x) = (f1(4) + x); a = f2(2)\n')
    with pytest.raises(AssertionError):  # fixme
        meta_model.model_from_str(inp)


def test_function_call_with_dummy_reference(meta_model):
    """test a call where expr is a reference to a dummy variable"""
    inp = 'f(x) = x; a = f(1); print(a)\n'
    prog = meta_model.model_from_str(inp)
    assert prog.value == str(1)


def test_call_undefined_function(meta_model):
    """test call of undefined function"""
    prog_inp = 'a = 2.*func()\n'
    with pytest.raises(TextXSemanticError,
                       match='Unknown object "func" of class "Function"'):
        meta_model.model_from_str(prog_inp)


def test_dummy_variable_names_in_different_scopes(meta_model):
    """test dummy variable with same names defined in different scopes"""
    prog_inp = 'sqr(x) = x*x; diff(x, y) = x - y; a = sqr(2.); b = diff(2, 1)\n'
    prog = meta_model.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    var_a = next(v for v in var_list if v.name == 'a')
    assert var_a.type_ == float
    assert var_a.value == 4.
    var_b = next(v for v in var_list if v.name == 'b')
    assert var_b.type_ == int
    assert var_b.value == 1


def test_duplicated_function_names(meta_model):
    """test repeated initializations"""
    prog_inps = ['use math.pi; pi = 4\n', 'v = 7.5; v(x, y) = x*x + y*y\n']
    for inp in prog_inps:
        with pytest.raises(InitializationError, match='repeated initialization'):
            meta_model.model_from_str(inp)


def test_function_call_invalid_number_of_parameters(meta_model):
    """test a function call with invalid number of parameters"""
    inps = ['f(x) = x*x; a = f()\n', 'f(x) = x*x; a = f(2, 3)\n']
    for inp in inps:
        with pytest.raises(TypeError, match='parameters but'):
            meta_model.model_from_str(inp)


def test_print_function_call(meta_model):
    """test a print of a function call"""
    inp = 'f(x) = x*x; print(f(3))\n'
    assert meta_model.model_from_str(inp).value == '9'


def test_function_definition_repeating_argument_names(meta_model):
    """test a function definition with repeating argument names"""
    inp = 'h(x, x) = x\n'
    # fixme: raise a more specific exception
    with pytest.raises(AssertionError):
        meta_model.model_from_str(inp)


def test_function_definition_with_call_dummies_equal_names(meta_model):
    """test a function definition with call dummies with repeating names"""
    inp = 'h(x, y) = x + y; g(x, y, z) = z + h(x, y)'
    # fixme: should work, improve substitution
    with pytest.raises(AssertionError):
        meta_model.model_from_str(inp)


def test_duplicated_dummy_variable_names_local_scope(meta_model):
    """test duplicated dummy variable names using local scope"""
    inp = 'sqr(x) = x*x; diff(x, y) = x - y; print(diff(2, 1))\n'
    assert meta_model.model_from_str(inp).value == '1'


def test_duplicated_names_global_and_local_scope(meta_model):
    """test repeating names in local and global scope"""
    inp = 'a = 2; f(a) = a + 1; c = f(a); print(c)\n'
    prog = meta_model.model_from_str(inp)
    assert prog.value == '3'
