"""
tests for the typechecker
"""
import os
import pytest
from textx import metamodel_from_file
from textx import get_children_of_type
from textx.exceptions import TextXSyntaxError
from constraints import ExpressionTypeError, TypeMismatchError
import metamodel
import constraints


@pytest.fixture(name='meta_model')
def fixture_metamodel():
    """parse the grammar and generate the object classes"""
    this_folder = os.path.abspath(os.path.dirname(__file__))
    mm_file = os.path.join(this_folder, '../grammar/virtmat.tx')
    meta = metamodel_from_file(mm_file, auto_init_attributes=False)
    metamodel.add_properties(meta)
    metamodel.add_processors(meta)
    constraints.add_processors(meta)
    return meta


def test_type_int_expression(meta_model):
    """test int type checker for arithmetic expressions and variables"""
    prog_str = "a = 10; b = -1 - a*3 + 15*(a-7)*(3+a) + 4*a*a"
    prog = meta_model.model_from_str(prog_str)
    var_objs = get_children_of_type('Variable', prog)
    exp_objs = get_children_of_type('Expression', prog)
    assert len(var_objs) == 2
    assert len(exp_objs) == 4
    for obj in var_objs:
        assert obj.type_ == int
    for obj in exp_objs:
        assert obj.type_ == int


def test_type_float_expression(meta_model):
    """test float type checker for arithmetic expressions and variables"""
    prog_str = "a = 10/c; b = 5*(8*a-3)/(15*(a-7.5)*(3+a)+4.1*a*a-a+6); c = 2"
    prog = meta_model.model_from_str(prog_str)
    var_objs = get_children_of_type('Variable', prog)
    exp_objs = get_children_of_type('Expression', prog)
    assert len(var_objs) == 3
    assert len(exp_objs) == 7
    var = next(v for v in var_objs if v.name == 'a')
    assert var.type_ == float
    var = next(v for v in var_objs if v.name == 'b')
    assert var.type_ == float
    var = next(v for v in var_objs if v.name == 'c')
    assert var.type_ == int
    for obj in exp_objs[:-1]:
        assert obj.type_ == float


def test_type_boolean_expression(meta_model):
    """test bool type checker for boolean expression objects and variables"""
    prog_str = "c = false; a = not (b and c); b = (c or true)"
    prog = meta_model.model_from_str(prog_str)
    var_objs = get_children_of_type('Variable', prog)
    exp_objs = get_children_of_type('Or', prog)
    assert len(var_objs) == 3
    assert len(exp_objs) == 5
    for obj in var_objs:
        assert obj.type_ == bool
    for obj in exp_objs:
        assert obj.type_ == bool


def test_type_string_objects(meta_model):
    """test str type checker for string objects and variables"""
    prog_str = "a = b; b = 'Abc'"
    prog = meta_model.model_from_str(prog_str)
    var_objs = get_children_of_type('Variable', prog)
    str_objs = get_children_of_type('String', prog)
    assert len(var_objs) == 2
    assert len(str_objs) == 1
    for obj in var_objs:
        assert obj.type_ == str
    for obj in str_objs:
        assert obj.type_ == str


def test_type_comparison_objects(meta_model):
    """test bool type checker for comparison objects and variables"""
    prog_str = "c = a > b; a = 1; b = 2"
    prog = meta_model.model_from_str(prog_str)
    var_objs = get_children_of_type('Variable', prog)
    cmp_objs = get_children_of_type('GreaterThan', prog)
    assert len(var_objs) == 3
    assert len(cmp_objs) == 1
    var = next(v for v in var_objs if v.name == 'c')
    assert var.type_ == bool


def test_expression_type_error(meta_model):
    """test expression type error catched after parsing"""
    prog_strs = ["bar = true; mixed_0 = 3 * bar # ExpressionTypeError",
                 "bar = true; threebar = 3*bar # ExpressionTypeError",
                 "notbar = false; twobar = 2*notbar # ExpressionTypeError"]
    for prog_str in prog_strs:
        with pytest.raises(ExpressionTypeError):
            meta_model.model_from_str(prog_str)


def test_expression_type_error_anonymous(meta_model):
    """test expression type error for expressions not assigned to variables"""
    prog_strs = ["string = 'Abc'; print(3+string)",
                 "foo = false; string = 'Abc'; print(foo+string)",
                 "foo = true; string = 'Abc'; print(foo*string)",
                 "foo = true; c = 4.3e-1; print(foo-c)",
                 "a = 10; c = 2.; print((a or c))",
                 "a = 1; c = -3; print((a and c))",
                 "a = -1e-5; c = 2.; print(not (a and c))"]
    for prog_str in prog_strs:
        with pytest.raises(ExpressionTypeError):
            meta_model.model_from_str(prog_str)


def test_type_mismatch_error(meta_model):
    """test type mismatch error catched after parsing"""
    prog_strs = ["twotypes = 3 if true else 0.5 # TypeMismatchError",
                 "twotypes = if(true, 3, 0.5) # TypeMismatchError"]
    for prog_str in prog_strs:
        with pytest.raises(TypeMismatchError):
            meta_model.model_from_str(prog_str)


def test_textx_syntax_error(meta_model):
    """test type errors catched during parsing"""
    prog_strs = ["bar = true; twobar = 2*(not bar)",
                 "bar = true; foo = false; mixed_2 = 'bar' + 'foo'",
                 "mixed_3 = 3 + 'foo'"]
    for prog_str in prog_strs:
        with pytest.raises(TextXSyntaxError):
            meta_model.model_from_str(prog_str)


def test_textx_semantic_error(meta_model):
    """test type errors catched during parsing"""
    prog_strs = ["mixed_1 = 3 * true # parser error",
                 "mixed_4 = 3 * null # parser error"]
    for prog_str in prog_strs:
        with pytest.raises(TextXSyntaxError):
            meta_model.model_from_str(prog_str)


def test_unbound_import_type_comparision(meta_model):
    """unbound import type in comparison expression """
    prog_inp = 'use exp from math; a = (exp(1) < exp(2))\n'
    prog = meta_model.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    calls = get_children_of_type('FunctionCall', prog)
    assert len(calls) == 2
    assert all(call.type_ is None for call in calls)
    var_a = next(v for v in var_list if v.name == 'a')
    assert var_a.type_ == bool


def test_import_type_comparision(meta_model):
    """inferred import type in comparison expression"""
    prog_inp = 'use exp from math; a = (exp(2) > 1); b = (exp(2) > 1.)\n'
    prog = meta_model.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    calls = get_children_of_type('FunctionCall', prog)
    assert len(calls) == 2
    assert all(call.type_ == (int, float) for call in calls)
    assert all(var.type_ == bool for var in var_list)


def test_unbound_import_type_if_function(meta_model):
    """unbound import type in if function"""
    prog_inp = 'use exp from math; a = if(true, exp(1), exp(2))\n'
    prog = meta_model.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    calls = get_children_of_type('FunctionCall', prog)
    assert len(calls) == 2
    assert all(call.type_ is None for call in calls)
    var_a = next(v for v in var_list if v.name == 'a')
    assert var_a.type_ is None


def test_import_type_if_function(meta_model):
    """inferred import type in if function"""
    prog_inp = 'use exp from math; a = if(true, exp(1), 3.1)\n'
    prog = meta_model.model_from_str(prog_inp)
    var_list = get_children_of_type('Variable', prog)
    calls = get_children_of_type('FunctionCall', prog)
    assert len(calls) == 1
    assert calls[0].type_ == float
    var_a = next(v for v in var_list if v.name == 'a')
    assert var_a.type_ == float


def test_function_call_type_in_print(meta_model):
    """bug function call type in print builtin"""
    inp = 'f(x) = 1 + x; print(f(1))\n'
    prog = meta_model.model_from_str(inp)
    calls = get_children_of_type('FunctionCall', prog)
    assert calls[0].type_ == int


def test_function_call_invalid_parameter_type(meta_model):
    """"test function calls with invalid parameter types"""
    inps = ['f(x) = x*x; a = f(true)\n', 'f(x) = if(x, 1, 0); a = f(1.)\n']
    for inp in inps:
        with pytest.raises(TypeError, match='from FunctionCall'):
            meta_model.model_from_str(inp)


def test_function_call_variable_type(meta_model):
    """test a function call with a variable type"""
    inp = 'f(x) = x; a = f(1); b = f(not false); c = f("Abc"); d = f(1.2e-5)\n'
    prog = meta_model.model_from_str(inp)
    var_list = get_children_of_type('Variable', prog)
    type_list = (int, bool, str, float)
    name_list = ('a', 'b', 'c', 'd')
    for name, typ in zip(name_list, type_list):
        var = next(v for v in var_list if v.name == name)
        assert var.type_ == typ
